#!/usr/bin/python

from threading import Thread
import time
import BaseHTTPServer
import SocketServer
import socket
import sys
import asyncore, asynchat
import re
from blogfish_functions import debug

PM = None

class ThreadingHTTPServer(SocketServer.ThreadingMixIn, BaseHTTPServer.HTTPServer):
	pass	
	
class FishHandler(BaseHTTPServer.BaseHTTPRequestHandler):

	def do_POST(self):
		body = self.rfile.read(int(self.headers.getheader("content-length")))
		debug("got: " + body[0:20] + " ...")
		self.send_response(200, "Got your message")
		self.send_header("Connection", "keep-alive")
		self.end_headers()
		# this never returns for JOIN PASV: the 
		# thread hangs around and becomes a PassiveFishTPClient
		# so the wfile remains open for sending to passive servers
		global PM
		PM.handle_message(body, self.client_address[0], self.wfile)
	
class FishTPServerBase(Thread):
	def __init__(self, port, pm):
		Thread.__init__(self)
		self.stop_flag = False
		self.port = port
		self.PM = pm
		global PM
		PM = self.PM
		self.setDaemon(1)

	def attempt_join(self, addr):
		pass

class FishTPServer(FishTPServerBase):
	def run(self):
		debug("PM is " + repr(self.PM))
		try:
			debug("starting to serve on port %s" % self.port)
			self.httpd = ThreadingHTTPServer(("127.0.0.1", int(self.port)), \
				FishHandler)
		except:
			debug("Couldn't start serving on port %s" % self.port)
			raise
		while (self.stop_flag == False):
		    self.httpd.handle_request()
				

class FishTPServerPassive(FishTPServerBase):
		
	def __init__(self, *args):
		FishTPServerBase.__init__(self, *args)
		self.map = {}
		self.lost_conns = {}
		
	def run(self):
		debug("starting passive server")
		while self.stop_flag == False:
			asyncore.loop(5, self.map)
			
	def attempt_join(self, addr):
		if addr in self.map: return True # nothing to do
		try:
			disp = FishTPDispatcher(self, addr)
		except:
			debug("couldn't create dispatcher")
			return
		self.map[addr] = disp
		if addr not in self.lost_conns: self.lost_conns[addr] = 0
		
	def got_message(self, message, host):
		self.PM.handle_message(message, host)


	def lost_connection(self, addr):
		debug("lost_connection to %s " % addr)
		if addr in self.map: del self.map[addr]
		if addr in self.lost_conns:
			self.lost_conns[addr] = self.lost_conns[addr] + 1
			if self.lost_conns[addr] > 2:
				del self.lost_conns[addr]
				self.PM.handle_failure(addr)
				return # don't keep trying to rejoin...
		self.attempt_join(addr) # temp

class FishTPDispatcher(asynchat.async_chat):

	def __init__(self, server, addr):
		asynchat.async_chat.__init__(self)
		self.server = server
		self.addr = addr
		self.set_terminator(None)
		self.mybuf = ""
		self.pat = re.compile(r"MESG (\d+)\n")
		host, port = addr.split(":", 2)
		
		try: # is this useful?
			self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
			self.connect((host, int(port)))
			# HACKTASTICA!!!
			body = "JOIN PASV\n"
			text = ("POST / HTTP/1.1\nContent-Length: %s\nContent-Type: " \
					+ "text/plain\nConnection: keep-alive\n\n") \
					% ( len(body))
			text = text + body
			self.push(text)
		except:
			debug("Failed to init asynchat channel")
			raise

	def collect_incoming_data(self, data):
		self.mybuf = self.mybuf + data
		while True:
			match = self.pat.search(self.mybuf)
			if not match: break
			end = match.end(0)
			bytes = int(match.group(1))
			if len(self.mybuf) < end + bytes: break
			message = self.mybuf[end:end + bytes]
			debug("Found message: '%s'" % message)
			self.server.got_message(message, self.addr.split(":")[0])
			self.mybuf = self.mybuf[end + bytes + 1:]

	def found_terminator(self):
		pass

	def handle_connect(self):
		pass

	def handle_write(self):
		self.push("\n")

	def handle_close(self):
		debug("handle_close")
		asynchat.async_chat.handle_close(self)
		self._lost_connection()

	def handle_error(self):
		# asynchat.async_chat.handle_error(self)
		self.close()
		self._lost_connection()

	def handle_expt(self):
		self._lost_connection()

	def _lost_connection(self):
		self.server.lost_connection(self.addr)

